@@ -19,7 +19,9 @@ class TripletLoss(GroupLoss):
     Args:
         margin: Margin value to push negative examples
             apart. Optional, defaults to `0.5`.
-        distance_metric_name: Name of the distance function, e.g., :class:`~quaterion.distances.Distance`. Optional, defaults to :attr:`~quaterion.distances.Distance.COSINE`.
+        distance_metric_name: Name of the distance function, e.g.,
+            :class:`~quaterion.distances.Distance`. Optional, defaults to
+            :attr:`~quaterion.distances.Distance.COSINE`.
         mining (str, optional): Triplet mining strategy. One of
             `"all"`, `"hard"`. Defaults to `"hard"`.
     """
