@@ -1,13 +1,18 @@
-from typing import Optional
+from typing import Optional, Union, Sized, Iterable
 
 import pytorch_lightning as pl
+from torch.utils.data import Dataset
+from quaterion_models import MetricModel
 
 from quaterion.dataset.similarity_data_loader import (
     PairsSimilarityDataLoader,
     GroupSimilarityDataLoader,
     SimilarityDataLoader,
 )
+from quaterion.eval.evaluator import Evaluator
 from quaterion.loss import GroupLoss, PairwiseLoss
+from quaterion.train.cleanup_callback import CleanupCallback
+from quaterion.train.metrics_callback import MetricsCallback
 from quaterion.train.trainable_model import TrainableModel
 
 
@@ -21,6 +26,7 @@ class Quaterion:
         trainer: pl.Trainer,
         train_dataloader: SimilarityDataLoader,
         val_dataloader: Optional[SimilarityDataLoader] = None,
+        ckpt_path: Optional[str] = None,
     ):
         """Handle training routine
 
@@ -34,6 +40,9 @@ class Quaterion:
                 stage
             val_dataloader: Optional DataLoader instance to retrieve samples during
                 validation stage
+            ckpt_path: Path/URL of the checkpoint from which training is resumed. If there is
+                no checkpoint file at the path, an exception is raised. If resuming from mid-epoch checkpoint,
+                training will start from the beginning of the next epoch.
         """
 
         if isinstance(train_dataloader, PairsSimilarityDataLoader):
@@ -49,12 +58,15 @@ class Quaterion:
                     "Try other loss/data loader"
                 )
 
+        trainer.callbacks.append(CleanupCallback())
+        trainer.callbacks.append(MetricsCallback())
+        # Prepare data loaders for training
+
         trainable_model.setup_dataloader(train_dataloader)
+        if val_dataloader:
+            trainable_model.setup_dataloader(val_dataloader)
 
-        if val_dataloader is not None:
-            val_dataloader.collate_fn = train_dataloader.collate_fn
-
-        trainable_model.cache(
+        trainable_model.setup_cache(
             trainer=trainer,
             train_dataloader=train_dataloader,
             val_dataloader=val_dataloader,
@@ -64,6 +76,14 @@ class Quaterion:
             model=trainable_model,
             train_dataloaders=train_dataloader,
             val_dataloaders=val_dataloader,
+            ckpt_path=ckpt_path,
         )
 
-        trainable_model.unwrap_cache()
+    @classmethod
+    def evaluate(
+        cls,
+        evaluator: Evaluator,
+        dataset: Union[Sized, Iterable, Dataset],
+        model: MetricModel,
+    ):
+        return evaluator.evaluate(dataset, model)
