@@ -1,3 +1,5 @@
+import os
+
 from typing import (
     Union,
     Dict,
@@ -10,7 +12,7 @@ from loguru import logger
 from quaterion_models.encoders import Encoder
 from torch.utils.data import DataLoader
 
-from quaterion.train.cache.cache_train_collater import CacheTrainCollater
+from quaterion.train.cache.cache_train_collater import CacheTrainCollator
 from quaterion.dataset.similarity_data_loader import SimilarityDataLoader
 from quaterion.train.cache import (
     CacheConfig,
@@ -20,6 +22,7 @@ from quaterion.train.cache import (
 )
 from quaterion.train.cache.cache_encoder import CacheMode
 from quaterion.train.cache.cache_model import CacheModel
+from quaterion.dataset.label_cache_dataset import LabelCacheMode
 
 
 class CacheMixin:
@@ -27,7 +30,7 @@ class CacheMixin:
     def _apply_cache_config(
         cls,
         encoders: Union[Encoder, Dict[str, Encoder]],
-        cache_config: Optional[CacheConfig],
+        cache_config: CacheConfig,
     ) -> Union[Encoder, Dict[str, Encoder]]:
         """Applies received cache configuration for cached encoders, remain
         non-cached encoders as is
@@ -50,7 +53,7 @@ class CacheMixin:
             ValueError: if CacheConfig instance does not have some of required
                 options set. E.g. not `mapping` nor `cache_type` being set
         """
-        if not cache_config:
+        if cache_config.cache_type == CacheType.NONE:
             return encoders
 
         if not cache_config.cache_type and not cache_config.mapping:
@@ -97,6 +100,9 @@ class CacheMixin:
             ValueError: if encoder layers are not frozen. Cache can be applied
                 only to fully frozen encoders' outputs.
         """
+        if isinstance(encoder, CacheEncoder):
+            return encoder
+
         if encoder.trainable:
             if encoder_name in cache_config.mapping:
                 raise ValueError(
@@ -126,17 +132,21 @@ class CacheMixin:
         train_dataloader: SimilarityDataLoader,
         val_dataloader: Optional[SimilarityDataLoader],
         cache_config: CacheConfig,
-    ) -> None:
+    ) -> bool:
         """Filling cache for model's cache encoders.
 
         Args:
             trainer: Lightning Trainer holds required parameters for model launch (gpu, e.t.c.)
-            encoders: mapping of model's encoders and their names
+            encoders: mapping of all model's encoders and their names
             train_dataloader: model's train dataloader
             val_dataloader: model's val dataloader
             cache_config: cache config instance to configure cache batch size
                 and num of workers to use for caching
 
+        Returns:
+            True, if cache was filled with data
+            False, if cache is not used
+
         """
         cache_encoders = {
             name: encoder
@@ -145,7 +155,15 @@ class CacheMixin:
         }
 
         if not cache_encoders:
-            return
+            return False
+
+        # Check if all encoders are cachable, and we don't use custom key extractor.
+        # If so, we can also cache whole dataset and avoid reading from it
+        is_full_cache_possible = (
+            len(cache_encoders) == len(encoders) and not cache_config.key_extractors
+        )
+        if is_full_cache_possible:
+            logger.debug("Using full cache")
 
         if cache_config.key_extractors and not isinstance(
             cache_config.key_extractors, dict
@@ -157,43 +175,65 @@ class CacheMixin:
         else:
             key_extractors = cache_config.key_extractors
 
-        cache_collater = CacheTrainCollater(
+        cache_collater = CacheTrainCollator(
             pre_collate_fn=train_dataloader.pre_collate_fn,
             encoder_collates={
                 name: encoder.get_collate_fn() for name, encoder in encoders.items()
             },
             key_extractors=key_extractors,
             cachable_encoders=list(cache_encoders.keys()),
-            mode=CacheMode.FILL,
+            mode=CacheMode.TRAIN,
         )
 
         train_dataloader.collate_fn = cache_collater
-
-        cache_train_dataloader = cls._wrap_cache_dataloader(
-            dataloader=train_dataloader, cache_config=cache_config
-        )
-
-        cache_val_dataloader = None
         if val_dataloader is not None:
             val_dataloader.collate_fn = cache_collater
-            cache_val_dataloader = cls._wrap_cache_dataloader(
-                dataloader=val_dataloader, cache_config=cache_config
-            )
 
-        cls._fill_cache(
-            trainer, cache_encoders, cache_train_dataloader, cache_val_dataloader
+        # Setup different cache key salt for train and val
+        train_dataloader.set_salt("train")
+        if val_dataloader is not None:
+            val_dataloader.set_salt("val")
+
+        is_persisted = cls.check_encoders_persisted(
+            cache_config.save_dir, cache_encoders
         )
 
-        cache_collater.mode = CacheMode.TRAIN
-        logger.info("Caching has been successfully finished")
+        if not is_persisted:
+            if is_full_cache_possible:
+                cls._label_cache_train_mode(train_dataloader, val_dataloader)
+
+            cache_collater.mode = CacheMode.FILL
+            cls._fill_cache(
+                trainer=trainer,
+                cache_encoders=cache_encoders,
+                train_dataloader=train_dataloader,
+                val_dataloader=val_dataloader,
+                cache_config=cache_config,
+            )
+            cache_collater.mode = CacheMode.TRAIN
+            logger.debug("Caching has been successfully finished")
+            cls.save_cache(
+                cache_config.save_dir, cache_encoders, train_dataloader, val_dataloader
+            )
+
+        else:
+            cls.load_cache(
+                cache_config.save_dir, cache_encoders, train_dataloader, val_dataloader
+            )
+
+        if is_full_cache_possible:
+            cls._enable_label_cache(train_dataloader, val_dataloader)
+
+        return True
 
     @classmethod
     def _fill_cache(
         cls,
         trainer: pl.Trainer,
         cache_encoders: Dict[str, CacheEncoder],
-        train_dataloader: DataLoader,
-        val_dataloader: DataLoader,
+        train_dataloader: SimilarityDataLoader,
+        val_dataloader: SimilarityDataLoader,
+        cache_config: CacheConfig,
     ) -> None:
         """Fills cache and restores trainer state for further training process.
 
@@ -204,12 +244,23 @@ class CacheMixin:
             val_dataloader: model's val dataloader
 
         """
+        cache_train_dataloader = cls._wrap_cache_dataloader(
+            dataloader=train_dataloader, cache_config=cache_config
+        )
+
+        cache_val_dataloader = None
+        if val_dataloader is not None:
+            cache_val_dataloader = cls._wrap_cache_dataloader(
+                dataloader=val_dataloader, cache_config=cache_config
+            )
+
         # The actual caching
         trainer.predict(
             CacheModel(
                 cache_encoders,
             ),
-            [train_dataloader, val_dataloader],
+            [cache_train_dataloader, cache_val_dataloader],
+            return_predictions=False,
         )
 
     @classmethod
@@ -261,3 +312,88 @@ class CacheMixin:
             dataset=dataloader.dataset, collate_fn=dataloader.collate_fn, **params
         )
         return cache_dl
+
+    @classmethod
+    def _label_cache_train_mode(
+        cls,
+        train_dataloader: SimilarityDataLoader,
+        val_dataloader: Optional[SimilarityDataLoader],
+    ):
+        train_dataloader.set_label_cache_mode(LabelCacheMode.learn)
+        if val_dataloader:
+            val_dataloader.set_label_cache_mode(LabelCacheMode.learn)
+
+    @classmethod
+    def _enable_label_cache(
+        cls,
+        train_dataloader: SimilarityDataLoader,
+        val_dataloader: Optional[SimilarityDataLoader],
+    ):
+        train_dataloader.set_skip_read(True)
+        train_dataloader.set_label_cache_mode(LabelCacheMode.read)
+        if val_dataloader:
+            val_dataloader.set_skip_read(True)
+            val_dataloader.set_label_cache_mode(LabelCacheMode.read)
+
+    @classmethod
+    def _encoders_cache_path(cls, dir_path: Optional[str]):
+        return os.path.join(dir_path, "encoders") if dir_path else None
+
+    @classmethod
+    def check_encoders_persisted(
+        cls, dir_path: Optional[str], encoders: Dict[str, Encoder]
+    ):
+        if not dir_path:
+            return False
+        encoders_path = cls._encoders_cache_path(dir_path)
+        for key, encoder in encoders.items():
+            if not os.path.exists(os.path.join(encoders_path, key)):
+                return False
+        return True
+
+    @classmethod
+    def save_cache(
+        cls,
+        dir_path: Optional[str],
+        encoders: Dict[str, Encoder],
+        train_dataloader: SimilarityDataLoader,
+        val_dataloader: Optional[SimilarityDataLoader],
+    ):
+        if not dir_path:
+            return
+        encoders_path = cls._encoders_cache_path(dir_path)
+        os.makedirs(encoders_path, exist_ok=True)
+        for key, encoder in encoders.items():
+            if isinstance(encoder, CacheEncoder):
+                encoder.save_cache(os.path.join(encoders_path, key))
+
+        train_dataloader.save_label_cache(os.path.join(dir_path, "train_labels"))
+        if val_dataloader:
+            val_dataloader.save_label_cache(os.path.join(dir_path, "val_labels"))
+        logger.debug(f"Cache saved to {dir_path}")
+
+    @classmethod
+    def load_cache(
+        cls,
+        dir_path: str,
+        encoders: Dict[str, Encoder],
+        train_dataloader: SimilarityDataLoader,
+        val_dataloader: Optional[SimilarityDataLoader],
+    ):
+        if not dir_path:
+            return
+        encoders_path = cls._encoders_cache_path(dir_path)
+        for key, encoder in encoders.items():
+            if isinstance(encoder, CacheEncoder):
+                encoder_cache_path = os.path.join(encoders_path, key)
+                if not os.path.exists(encoder_cache_path):
+                    raise RuntimeError(
+                        "Encoder cache was configured, but not found. "
+                        f"Expected to find cache at {encoder_cache_path}, but file does not exists!"
+                    )
+                encoder.load_cache(encoder_cache_path)
+
+        train_dataloader.load_label_cache(os.path.join(dir_path, "train_labels"))
+        if val_dataloader:
+            val_dataloader.load_label_cache(os.path.join(dir_path, "val_labels"))
+        logger.debug(f"Cache loaded from: {dir_path}")
