@@ -1,11 +1,8 @@
 from __future__ import annotations
-from typing import Dict, Any, Union, Optional
+from typing import Dict, Any, Union, Optional, List
 
 import pytorch_lightning as pl
-from pytorch_lightning.utilities.types import (
-    TRAIN_DATALOADERS,
-    EVAL_DATALOADERS,
-)
+
 from quaterion_models import MetricModel
 from quaterion_models.encoders import Encoder
 from quaterion_models.heads import EncoderHead
@@ -13,10 +10,12 @@ from quaterion_models.types import TensorInterchange
 from torch import Tensor
 
 from quaterion.dataset import SimilarityDataLoader
-from quaterion.dataset.train_collater import TrainCollater
+from quaterion.dataset.train_collater import TrainCollator
+from quaterion.eval.attached_metric import AttachedMetric
 from quaterion.loss import SimilarityLoss
 from quaterion.train.cache import (
     CacheConfig,
+    CacheType,
 )
 from quaterion.train.cache_mixin import CacheMixin
 from quaterion.utils.enums import TrainStage
@@ -29,14 +28,74 @@ class TrainableModel(pl.LightningModule, CacheMixin):
         super().__init__(*args, **kwargs)
 
         encoders = self.configure_encoders()
-        self.cache_config = self.configure_caches()
-        encoders = self._apply_cache_config(encoders, self.cache_config)
+        self.cache_config: CacheConfig = self.configure_caches() or CacheConfig(
+            cache_type=CacheType.NONE
+        )
 
         head = self.configure_head(MetricModel.get_encoders_output_size(encoders))
 
+        metrics = self.configure_metrics()
+        self.attached_metrics: List[AttachedMetric] = (
+            [metrics] if isinstance(metrics, AttachedMetric) else metrics
+        )
+
         self._model = MetricModel(encoders=encoders, head=head)
         self._loss = self.configure_loss()
 
+    def configure_metrics(self) -> Union[AttachedMetric, List[AttachedMetric]]:
+        """Method to configure batch-wise metrics for a training process
+
+        Use this method to attach batch-wise metrics to a training process.
+        Provided metrics have to have similar to
+        :class:`~quaterion.eval.pair.pair_metric.PairMetric` or
+        :class:`~quaterion.eval.group.group_metric.GroupMetric`
+
+        Returns:
+            Union[:class:`~quaterion.eval.attached_metric.AttachedMetric`,
+            List[:class:`~quaterion.eval.attached_metric.AttachedMetric`]] -
+            metrics attached to the model
+
+        Examples::
+
+            return [
+                AttachedMetric(
+                    "RetrievalPrecision",
+                    RetrievalPrecision(k=1),
+                    prog_bar=True,
+                    on_epoch=True,
+                ),
+                AttachedMetric(
+                    "RetrievalReciprocalRank",
+                    RetrievalReciprocalRank(),
+                    prog_bar=True,
+                ),
+            ]
+        """
+        return []
+
+    def _evaluate(
+        self,
+        embeddings: Tensor,
+        targets: Dict[str, Any],
+        stage: TrainStage,
+    ) -> None:
+        """Method to calculate and log metrics, accumulate embeddings in estimators
+
+        Calculate current stage and batch metrics
+
+        Args:
+            embeddings: current batch embeddings
+            targets: objects to calculate labels for similarity samples
+            stage: training, validation, etc.
+        """
+        for metric in self.attached_metrics:
+            if stage in metric.stages:
+                self.log(
+                    f"{metric.name}_{stage}",
+                    metric.compute(embeddings, **targets),
+                    **metric.log_options,
+                )
+
     @property
     def model(self) -> MetricModel:
         """Origin model to be trained
@@ -149,7 +208,7 @@ class TrainableModel(pl.LightningModule, CacheMixin):
     def training_step(
         self, batch: TensorInterchange, batch_idx: int, **kwargs
     ) -> Tensor:
-        """Compute and return the training loss and some additional metrics for e.g.
+        """:meta private: Compute and return the training loss and some additional metrics for e.g.
         the progress bar or logger.
 
         Args:
@@ -167,7 +226,7 @@ class TrainableModel(pl.LightningModule, CacheMixin):
         return loss
 
     def validation_step(self, batch, batch_idx, **kwargs) -> Optional[Tensor]:
-        """Compute validation loss and some additional metrics for e.g. the progress
+        """:meta private: Compute validation loss and some additional metrics for e.g. the progress
         bar or logger.
 
         Args:
@@ -180,7 +239,7 @@ class TrainableModel(pl.LightningModule, CacheMixin):
         return None
 
     def test_step(self, batch, batch_idx, **kwargs) -> Optional[Tensor]:
-        """Compute test loss and some additional metrics for e.g. the progress
+        """:meta private: Compute test loss and some additional metrics for e.g. the progress
         bar or logger.
 
         Args:
@@ -206,9 +265,17 @@ class TrainableModel(pl.LightningModule, CacheMixin):
             Tensor: computed loss value
         """
         features, targets = batch
+
         embeddings = self.model(features)
         loss = self.loss(embeddings=embeddings, **targets)
         self.log(f"{stage}_loss", loss)
+
+        self._evaluate(
+            embeddings=embeddings,
+            targets=targets,
+            stage=stage,
+        )
+
         self.process_results(
             embeddings=embeddings,
             targets=targets,
@@ -216,6 +283,7 @@ class TrainableModel(pl.LightningModule, CacheMixin):
             stage=stage,
             **kwargs,
         )
+
         return loss
 
     def save_servable(self, path: str):
@@ -227,14 +295,32 @@ class TrainableModel(pl.LightningModule, CacheMixin):
         self.unwrap_cache()
         self.model.save(path)
 
-    def cache(
+    def setup_cache(
         self,
         trainer: pl.Trainer,
         train_dataloader: SimilarityDataLoader,
         val_dataloader: Optional[SimilarityDataLoader],
     ):
-        """Fill cachable encoders with embeddings"""
-        self._cache(
+        """:meta private: Prepares encoder's cache for faster training:
+
+        - Replaces frozen encoders with cache Wrapper according
+          to the cache configuration defined in :meth:`configure_caches`.
+        - Fill cachable encoders with embeddings
+
+        Args:
+            trainer: Pytorch Lightning trainer object
+            train_dataloader: train dataloader
+            val_dataloader: validation dataloader
+
+        Returns:
+            `False`, if cache is already created or not required.
+            `True`, if cache is newly created
+        """
+        self.model.encoders = self._apply_cache_config(
+            self.model.encoders, self.cache_config
+        )
+
+        return self._cache(
             trainer=trainer,
             encoders=self.model.encoders,
             train_dataloader=train_dataloader,
@@ -243,39 +329,23 @@ class TrainableModel(pl.LightningModule, CacheMixin):
         )
 
     def unwrap_cache(self):
-        """Restore original encoders"""
+        """:meta private: Restore original encoders"""
         self.model.encoders = self._unwrap_cache_encoders(self.model.encoders)
 
     def setup_dataloader(self, dataloader: SimilarityDataLoader):
-        """Update data loader's collate function with encoder-specific collate"""
+        """Setup data loader for encoder-specific settings, Setup encoder-specific collate function
+
+        Each encoder have its own unique way to transform a list of records into NN-compatible format.
+        These transformations are usually done during data pre-processing step.
+        """
         encoder_collate_fns = dict(
             (key, encoder.get_collate_fn())
             for key, encoder in self.model.encoders.items()
         )
 
-        collater = TrainCollater(
+        collator = TrainCollator(
             pre_collate_fn=dataloader.collate_fn,
             encoder_collates=encoder_collate_fns,
         )
 
-        dataloader.collate_fn = collater
-
-    # region anchors
-    # https://github.com/PyTorchLightning/pytorch-lightning/issues/10667
-    def train_dataloader(self) -> TRAIN_DATALOADERS:
-        """:meta private:"""
-        pass
-
-    def test_dataloader(self) -> EVAL_DATALOADERS:
-        """:meta private:"""
-        pass
-
-    def val_dataloader(self) -> EVAL_DATALOADERS:
-        """:meta private:"""
-        pass
-
-    def predict_dataloader(self) -> EVAL_DATALOADERS:
-        """:meta private:"""
-        pass
-
-    # endregion
+        dataloader.collate_fn = collator
